
## CODE IS RUN IN JUPYTER NOTEBOOK

%matplotlib inline

import numpy as np
import import_ipynb
import glob

from scipy.io import wavfile
import soundfile as sf
import librosa as lb
import librosa.display
import matplotlib.pyplot as plt

from scipy.spatial.distance import euclidean

from fastdtw import fastdtw



def read_filename():
    filename = input('Enter the full filename, make sure to take the path as well if the file is not in the same directory as this script. :')
    return filename


%%time
filename = read_filename()
sig, fs = lb.load(filename, sr = 16000) # Returns a 1D array with samples. For more than one channel, it returns [channels,samples]

import time
time_stamp = []
temp = [] # np.zeros(sig.size)

for sample in range(0,sig.size,64000):
    temp = np.append(temp, sample)
    time_stamp = np.append(time_stamp,sample/fs)



def time_to_index(timestamp):
    time = timestamp - 1000000 # removes the 1 in front of the timestamp
    sample = time*fs
    return sample


def convert(seconds): #converts to hh:min:sec
    return time.strftime("%H:%M:%S", time.gmtime(seconds))

################################################################################
from scipy.io.wavfile import write
import os

path = '/Users/Bruker/master20/SoundJS'
try:
    os.mkdir(path)
except OSError:
    print('failed')
else:
    print('succsess')

def give_file_name(index):
    t = convert(time_stamp[index])
    new_time = t.replace(':', '')
    old_name = filename[:filename.find('.')]
    new_name = path +'/'+ old_name + '_' + '!1' + new_time + '.wav'
    #new_name =  old_name + '_' + '!1' + new_time + '.wav'
    write(new_name, fs, sig[int(temp[index-1]):int(temp[index])])
    return new_name
################################################################################

# Write the timestamp of the data with corresponding confidence interval to a .txt file
def add_new_line(filename, start, end, prob):
    data = []
    with open(filename, "a+") as datafile_id:
        # Move read cursor to the start of file.
        datafile_id.seek(0)
        # If file is not empty then append '\n'
        data = datafile_id.read(100)
        if len(data) > 0:
            datafile_id.write("\n")
        # Add text to end of file
        data = np.array([start, end, prob])
        data = data.T

        # write to a .txt file, with tab as a delimiter to be abel to import the timestamps to Audacity
        np.savetxt(datafile_id, data, fmt='%s', delimiter = '\t')
################################################################################

def mel_spec():
    ps = lb.feature.melspectrogram(sig, sr=fs, n_fft = 2048 ) #power spectrum
    ps_dB = librosa.power_to_db(ps, ref=np.max)
    librosa.display.specshow(ps, y_axis='mel', x_axis='time', sr=fs, fmax = 1500)
    #librosa.display.specshow(ps, y_axis='mel', x_axis='time', sr=fs, fmax = 1500)
    plt.colorbar(format = '%+2.0f dB')
    plt.title('Mel-frequency spectrogram')


from scipy.signal import butter,lfilter, freqz

####################### Setup Butter bandpassfilter ############################
def butter_bandpass(lowcut, highcut, fs, order = 5):
    nyq = 0.5*fs
    low = lowcut/nyq
    high = highcut/nyq
    b, a = butter(order, [low,high], btype = 'band')
    #return
    return b, a

################################### BP filter ##################################
def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    b, a = butter_bandpass(lowcut, highcut, fs, order=order)
    y = lfilter(b, a, data)
    return y

################################################################################
def check_if_present(file, new_word):
    with open(file) as file_object:
        # Move read cursor to the start of file.
        content = file_object.read()
        if new_word in content:
            print('This sound clip is already detected and stored')
            return 0
        else:
            return 1
################################################################################
def store_timestamp(file1):
    timestamp = []
    # read through all files in directory
    # see which files where difference is 4 seconds

    file1_name = file1[:file1.find('.')] # Timestamp is now the last

    file1_nr = file1_name[file1_name.find('!'):]
    file1new = int(file1_nr.replace('!', '')) # særs innefektivt, finn raskere måte?
    timestamp = np.append(timestamp, file1new)
    return timestamp
################################################################################

# Gives the number in a string array on the first place. returned numb is a str
def split_str(array_place):
    old_str = array_place[:array_place.find(']')]
    old = old_str.replace('[','')
    old = old_str.replace('[','')
    m = old.split(' ')
    numb = m[0]
    return numb

################################################################################

def findMax(arr):
    # Initialize maximum element
    max = arr[0]
    # set max to beginning of array
    # update max as transversing through the array
    for i in range(len(arr)-1):
        if arr[i] > max:
            max = arr[i]
    return max

################################################################################

import copy
def preprocess_mfcc(mfcc):
    mfcc_cp = copy.deepcopy(mfcc)
    for i in range(mfcc.shape[1]):
        mfcc_cp[:,i] = mfcc[:,i] - np.mean(mfcc[:,i])
        mfcc_cp[:,i] = mfcc_cp[:,i]/np.max(np.abs(mfcc_cp[:,i]))
    return mfcc_cp


##############################################################################
                        Used when applying DTW
###############################################################################
def findMinDist(sound):
    min_dist = []

    y_dtw1, sr_dtw1 = lb.load('/Users/Bruker/master20/train/master/JS11.wav', sr = 16000) #filtered signal, 400-2000 Hz
    mfccJS1  = lb.feature.mfcc(y_dtw1)
    mfccJS1 = preprocess_mfcc(mfccJS1)

    y_dtw2, sr_dtw2 = lb.load('/Users/Bruker/master20/train/master/js_17.wav', sr = 16000) #filtered signal, 400-2000 Hz
    mfccJS2  = lb.feature.mfcc(y_dtw2)
    mfccJS2 = preprocess_mfcc(mfccJS2)

    y_dtw3, sr_dtw3 = lb.load('/Users/Bruker/master20/train/master/JS45.wav', sr = 16000) #filtered signal, 400-2000 Hz
    mfccJS3  = lb.feature.mfcc(y_dtw3)
    mfccJS3 = preprocess_mfcc(mfccJS3)

    y_dtw4, sr_dtw4 = lb.load('/Users/Bruker/master20/train/master/js_8.wav', sr = 16000) #filtered signal, 400-2000 Hz
    mfccJS4  = lb.feature.mfcc(y_dtw4)
    mfccJS4 = preprocess_mfcc(mfccJS4)

    JS_bpsound = butter_bandpass_filter(sound, 400, 2000, fs, order = 5)
    mfccJS_predicted = librosa.feature.mfcc(JS_bpsound,sr = 16000)
    mfccJS_predicted = preprocess_mfcc(mfccJS_predicted)

    # compares the input sound signal sound with 4 clean vocalizations of jack snipe
    distance1, path2 = fastdtw(mfccJS_predicted.T, mfccJS1.T)
    min_dist = np.append(min_dist, distance1)
    distance2, path3 = fastdtw(mfccJS_predicted.T, mfccJS2.T)
    min_dist = np.append(min_dist, distance2)
    distance3, path4 = fastdtw(mfccJS_predicted.T, mfccJS3.T)
    min_dist = np.append(min_dist, distance3)
    distance4, path5 = fastdtw(mfccJS_predicted.T, mfccJS4.T)
    min_dist = np.append(min_dist, distance4)

    distance = min(min_dist)
    return distance

################################################################################

################################################################################
        CLASSIFICATION OF THE INPUT RECORDING, WHERE THE MAGIC HAPPENS
################################################################################
%%time
# CURRENT


from keras.models import load_model
path = '/Users/Bruker/master20/SoundJS_CNN_90mfcc_3k16kHz' # Path to dir where soundfiles are saved

try:
    os.mkdir(path)
except OSError:
    print('Allready exsists')
else:
    print('succsess')
data = []
predict_sound = []
prob = []
start_array = []
end_array = []
prob_array = []
JS_predicted = []
timestamp = []
probability_all = ()
probability_js = ()
prob = []
model = load_model('best_CNN_model_90mfcc_16kHz.h5')
nr_mel = 90 # cHANGED DEPENDING ON THE INPUT THE MODEL IS TRAINED ON
dur_if = 0
dur_else = 0
dur_tot = 0

for i in range(len(temp)-1):
    sound = sig[int(temp[i]):int(temp[i+1])]
    sound = lb.util.normalize(sound)
    # Change between 2000 and 4000 dependng on the wanted frequency range
    f_sound = butter_bandpass_filter(sound, 400, 3000, fs, order = 5)

    # ps is the mel-spec that the model can use to predict
    # n_fft = 2048 corresponds to 128 ms for length of fft window
    ps = lb.feature.melspectrogram(y=f_sound, sr=fs, window = 'hann', n_fft = 2048, n_mels = nr_mel)

    #For CNN model, it should be three inputs (nr_mel, 126, 1), for LSTM it should be two (nr_mel, 126)
    ps_reshape = np.array([ps.reshape(nr_mel, 126, 1)])


    predict_sound = np.append(predict_sound, model.predict_classes(ps_reshape))

    #For CNN model, it should be four inputs (1, nr_mel, 126, 1), for LSTM it should be three (1, nr_mel, 126)
    probability_all = np.append(probability_all, str(model.predict([ps_reshape.reshape(1,nr_mel,126, 1)])))



    # 0 means JS positive, 1 means NJS positive
    # the probability array contains the probability for each clip for values between 0-1
#################### REMOVE WHEN USING DTW ####################################
    if predict_sound[i] == 0:
        JS_predicted = np.append(JS_predicted, temp[i]) # sample nr. of the predicted sound, replaces temp.
        timestamp = np.append(timestamp,temp[i]/fs) #timestamp of where JS vocalizes
        probability_js = np.append(probability_js, probability_all[i])
        # probability for the predicted JS clips

################ FOR DTW, UNCOMMENT THIS #######################################
#    if predict_sound[i] == 0:
#        distance = findMinDist(sound_raw)
#        if distance < 50:
#            JS_predicted = np.append(JS_predicted, temp[i]) # sample nr. of the predicted sound, replaces temp.
#            timestamp = np.append(timestamp,temp[i]/fs) #timestamp of where JS vocalizes
#            probability_js = np.append(probability_js, probability_all[i])


i = 0
while i in range(len(timestamp)-1):
    index = 4
    if timestamp[i] == (timestamp[i+1] - 4):
        start = timestamp[i]-4 # time in sec, needs to take -4 as hit is in endpoint to get start to startpoint
        start_array = np.append(start_array, start)
        i_start = JS_predicted[i] - 64000 # value of the sample
        no = i
        #print('Enter here', 'Index no:',i, no, 'Timestamp:',timestamp[i])
        while no < len(timestamp)-1 and int(timestamp[i]) == int(timestamp[no+1])-index:
            index = index+4
            prob = np.append(prob, split_str(probability_js[no]))
            no = no + 1
            i_end = JS_predicted[no]
            end = timestamp[no] # time in sec

        max_prob = findMax(prob)

        prob_array = np.append(prob_array, float(max_prob))
        end_array = np.append(end_array, end)

        JSsound_clip = sig[int(i_start):int(i_end)]
        duration = convert(end-start)
        dur_if = dur_if +  end-start
        t_s = convert(start)
        t_e = convert(end)

        new_p = max_prob.replace('.', '')
        new_s = t_s.replace(':', '')
        new_e = t_e.replace(':', '')
        duration = duration.replace(':', '')

        old_name = filename[:filename.find('.')]
        name_txt = old_name +  '_' + new_s + '_' + duration + '_' + new_p + '.wav'
        new_name = path +'/'+ old_name +  '_' + new_s + '_' + duration + '_' + new_p + '.wav'
        #print(new_name)

        #create .wav file with fs = 16 kHz,
        #with name including start time, duration and confidence interva
        write(new_name, fs, JSsound_clip)


        prob = []
        # To reset the arry so it does not take maximum of all the predictions

        i = no + 1

    else:
        #JS_predicted stores the last sample in a clip, so the start is -64000
        i_start = JS_predicted[i] - 64000

        i_end = JS_predicted[i]

        #To get the start point and not the end point, minus 4 as this corresponds to 64000
        # with fs = 16 kHz
        start = timestamp[i] - 4
        start_array = np.append(start_array, start)
        end_array = np.append(end_array, start+4)

        t_s = convert(start)
        duration = convert((i_end - i_start)/fs)
        dur_else = dur_else + (i_end-i_start)/fs
        p = split_str(probability_js[i])
        prob_array = np.append(prob_array, float(p))

        new_p = p.replace('.','')
        new_s = t_s.replace(':', '')
        duration = duration.replace(':', '')

        old_name = filename[:filename.find('.')]
        new_name = path +'/'+ old_name + '_' + new_s + '_' + duration + '_' + new_p + '.wav'

        write(new_name, fs, sig[int(i_start):int(i_end)])

        i = i + 1
    dur_tot = dur_if + dur_else
# write start and end time to a txt file, with corresponding confidence interval to be imported to Audacity
add_new_line('JS_soundDetected_CNN_90mfcc2_3k16kHz.txt', start_array, end_array, prob_array)
