# SCRIPT FOR TRAINING THE MODELS, FROM URBANSOUND CHALLENGE



%matplotlib inline
from memory_profiler import memory_usage
import os
from glob import glob
import matplotlib.pyplot as plt
import numpy as np
import h5py


import keras
# All the layers used to build the LSTM and CNN model
from keras.layers import Activation, Dense, Dropout, Conv2D, Flatten, MaxPooling2D, GaussianNoise
from keras.models import Sequential
from keras.utils.vis_utils import plot_model
from keras.metrics import AUC
from keras.callbacks import EarlyStopping,ReduceLROnPlateau,ModelCheckpoint,TensorBoard,ProgbarLogger
from keras import regularizers, optimizers
from sklearn.model_selection import train_test_split
from sklearn.utils import class_weight
from sklearn.metrics import f1_score, roc_curve, confusion_matrix, auc
import librosa
import librosa.display
import numpy as np
from sklearn.preprocessing import MultiLabelBinarizer
import random
import warnings
warnings.filterwarnings('ignore')

#object serialization
import _pickle as cPickle  #python 3 change, bruker jeg denne ?
import os
nr_mel = 90 # Change depending on the wanted mel bands in the filter bank for melspectrogram
%matplotlib inline

#enable garbage collection control
import gc
gc.enable()

print(h5py.__version__)


#progress tracker
from tqdm import tqdm, tqdm_notebook

# when set to TRUE, training data get loaded from a saved serialized data object file
# All audio files data get saved to a serialized object file to save reloading time on training runs
SKIP_AUDIO_RELOAD = False

#location of the sound files
INPUT_PATH='/Users/Bruker/master20'

TRAIN_INPUT=INPUT_PATH+'/train'
TRAIN_AUDIO_DIR=TRAIN_INPUT+'/master'

TEST_INPUT='/Users/Bruker/prosjektH19/test'
TEST_AUDIO_DIR=TEST_INPUT+'/Test'

def load_input_data(pd, filepath):
    # Read Data
    data = pd.read_csv(filepath)

# training file
TRAIN_FILE=TRAIN_INPUT+'/JS.csv'

#show info
train_input=load_input_data(pd,TRAIN_FILE)
train_input.head(8)

# test file
TEST_FILE=TEST_INPUT+'/test.csv'

#show info
test_input=load_input_data(pd,TEST_FILE)
test_input.head()


#labels
valid_train_label = train_input[['Class']]
valid_train_label.count()

#unique classes
x = train_input.groupby('Class')['Class'].count()
x

print(len(train_input[['Class']]))

# train data size
valid_train_data = train_input[['file_ID', 'Class']]
valid_train_data.count()

# test data size
valid_test_data = test_input[['file_ID']]
valid_test_data.count()

#training audio files
valid_train_data['path'] = TRAIN_AUDIO_DIR+'/' + train_input['file_ID'].astype('str')+".wav"
print ("sample",valid_train_data.path[1])
valid_train_data.head(5)

#test audio files
valid_test_data['path'] = TEST_AUDIO_DIR+'/' + test_input['file_ID'].astype('str') +".wav"
print ("sample",valid_test_data.path[1])

valid_test_data.head(5)


################################## LOADING AUDIO ###############################

# set duration on audio loading to make audio content to ensure each training
# data have same size
# fix the load audio file size
audio_play_duration= 4

def load_audio_file(file_path, duration=4, sr=16000):
    #load 4 seconds audio file, sr=22050
    # To preserve the native sampling rate of the file, use sr=None
    input_length=sr*duration
    # Load an audio file as a floating point time series.
    # y : np.ndarray [shape=(n,) or (2, n)] - audio time series
    # sr : number > 0 [scalar] - sampling rate of y
    y, sr = librosa.load(file_path,sr=sr, duration=duration)
    dur = librosa.get_duration(y=y)
    #pad output if audio file less than duration
    # Use edge-padding instead of zeros
    if (round(dur) < duration):
        offset = len(y) - round(input_length)
        print ("fixing audio length :", file_path)
        print ("input:", round(input_length), " load:", len(y) , " offset:", offset)
        y = librosa.util.fix_length(y, round(input_length))
        y = librosa.util.normalize(y)
    # using a pre-computed power spectrogram
    # Short-time Fourier transform (STFT)
   #D = np.abs(librosa.stft(y))**2
    #ps = librosa.feature.melspectrogram(S=D)    # ENDRE SR FOR Å SE OM DET KAN HJELPE
    #ps = librosa.feature.mfcc(y=y, sr=16000, n_mfcc=13)
    #ps = librosa.feature.melspectrogram(y=y, sr=16000, n_fft=1024, hop_length=512, win_length=None, window='hann', n_mels = nr_mel ) #alt bak sr er lagt på
    ps = librosa.feature.melspectrogram(y=y, sr=16000, window='hann', n_mels = nr_mel, )
    return ps

############################## RESHAPE TO DESIRED INPUT ########################
%%time
# Dataset
train_audio_data = []
train_object_file='saved_train_audio_data.p'

#override the reload flag if serized file doesn't exist
if not os.path.isfile(train_object_file):
    SKIP_AUDIO_RELOAD = False
127,207,
#load training data
if SKIP_AUDIO_RELOAD is True:
    print ("skip re-loading TRAINING data from audio files")
else:
    print ("loading train audio data, may take more than 1 minute. please wait!")
    for row in tqdm(valid_train_data.itertuples()):
            ps = load_audio_file(file_path=row.path, duration=4.0)
            if ps.shape != (nr_mel, 126): continue
            train_audio_data.append( (ps, row.Class) )
    print("Number of train samples: ", len(train_audio_data))
# have 400 samples, took 12.9 seconds to finish

################################################################################
# test data size
valid_test_data = test_input[['file_ID']]
valid_test_data.count()

# load saved data
test_audio_data = []
test_object_file='saved_test_audio_data.p'
if SKIP_AUDIO_RELOAD is True:
    test_audio_data = cPickle.load(open(test_object_file, 'rb'))
    print ("loaded test data [%s] records from object file" % len(test_audio_data))
else:
    cPickle.dump(test_audio_data, open(test_object_file, 'wb'))
    print ("save loaded test data :", len(test_audio_data))

#prepare for training
from sklearn.preprocessing import LabelEncoder
from sklearn.preprocessing import OneHotEncoder
from keras.utils import to_categorical
from numpy import argmax

# get a set of unique text labels
list_labels = sorted(list(set(valid_train_data.Class.values)))
print ("unique text labels count: ",len(list_labels))
print ("labels: ",list_labels)

# integer encode
label_encoder = LabelEncoder()
label_integer_encoded = label_encoder.fit_transform(list_labels)
print("encoded labelint values", label_integer_encoded)

# one hot encode
encoded_test = to_categorical(label_integer_encoded)
inverted_test = argmax(encoded_test[0])
#print(encoded_test, inverted_test)

#map filename to label
file_to_label = {k:v for k,v in zip(valid_train_data.path.values, valid_train_data.file_ID.values)}

# Map integer value to text labels
label_to_int = {k:v for v,k in enumerate(list_labels)}
print ("test label to int ",label_to_int["NJS"])

# map integer to text labels
int_to_label = {v:k for k,v in label_to_int.items()}
print ("int to label: ",int_to_label)

#full dataset
dataset = train_audio_data
random.shuffle(dataset)

# Split the training dataset into 0.7 training and 0.3 test
RATIO=0.7
train_cutoff= round(len(dataset) * RATIO)
train = dataset[:train_cutoff]
test = dataset[train_cutoff:]

X_train, y_train = zip(*train)
X_test, y_test = zip(*test)

# Reshape for CNN input
X_train = np.array([x.reshape( (nr_mel, 126, 1) ) for x in X_train])
X_test = np.array([x.reshape( (nr_mel, 126, 1) ) for x in X_test])

print ("train ",X_train.shape, len(y_train))
print ("test ", X_test.shape, len(y_test))


# Apply sck-learn label text encoding to integer
label_encoder = LabelEncoder()
y_train_integer_encoded = label_encoder.fit_transform(y_train)
y_test_integer_encoded = label_encoder.fit_transform(y_test)

print('y_test:', y_test_integer_encoded)




# Apply Keras One-Hot encoding for classes
y_train = np.array(keras.utils.to_categorical(y_train_integer_encoded, len(list_labels)))
y_test = np.array(keras.utils.to_categorical(y_test_integer_encoded, len(list_labels)))


#split up test into 0.5  test and 0.5 validation
X_test, X_val, y_test, y_val = train_test_split(X_test, y_test, test_size=0.5, random_state = 42)

print ("test set ",X_test.shape, len(y_test))
print ("validation set ", X_val.shape, len(y_val))


# In[29]:


# Apply weight to balance the classes
Class_Weights = class_weight.compute_class_weight('balanced', np.unique(y_train_integer_encoded), y_train_integer_encoded)
###############################################################################
            #    LSTM MODEL
############################## BUILDING THE MODEL ##############################
# FOR LSTM MODEL, CHANGE THE INPUT SIZE TO (N_MELS, 126), AND BUILD AS BELOW

# Build LTSM model, input shape = (nr_mel,  126)

#input_shape = X_train.shape[1:]

# model = Sequential()

# model.add(LSTM(8, input_shape=input_shape, return_sequences=True))
# model.add(Activation('tanh'))

# model.add(LSTM(16, return_sequences=True))
# model.add(Activation('tanh'))


# model.add(Flatten())
# model.add(Dropout(rate=0.5))

# model.add(Dense(16))
# model.add(Activation('tanh'))
# model.add(Dropout(rate=0.5))

# model.add(Dense(len(list_labels)))
# model.add(Activation('softmax'))

# model.summary()

################################################################################
                #    CNN MODEL
################################################################################
# build convolution model
# input shape = (nr_mels, 126, 1) for sr = 16 kHz
# model = Sequential()
# input_shape= X_train.shape[1:]

model.add(Conv2D(16, (3, 3), strides=(2, 2), input_shape= input_shape))
model.add(MaxPooling2D(2))
model.add(Activation('relu'))
################################### ADD AN EXTRA LAYER IF UNDERFITTED
#model.add(Conv2D(16, (3, 3), strides=(2, 2), padding="valid"))
#model.add(MaxPooling2D(2))
#model.add(Activation('relu'))
#model.add(Dropout(rate=0.5))
#######################################################################
model.add(Conv2D(24, (3, 3), strides=(2, 2), padding="valid"))
model.add(Activation('relu'))
model.add(Dropout(rate=0.3))

model.add(Flatten())
model.add(Dropout(rate=0.3))

model.add(Dense(24))
model.add(Activation('relu'))
model.add(Dropout(rate=0.3))

model.add(Dense(len(list_labels)))
model.add(Activation('softmax'))

model.summary()

########################### TRAINING OF THE MODEL ##############################
%%time

MAX_EPOCHS= 18 # INCREASE/DECREASE DEPENDENT ON UNDERFITTED/OVERFITTED MODEL
MAX_BATCH_SIZE= 32   # LARGER BATCH SIZE INCREASES ACCURACY
MAX_PATIENT= 2

# saved model checkpoint file
best_model_file="./best_CNN_model_90mfcc_16kHz_adam.h5"
# callbacks
callback=[ReduceLROnPlateau(patience=MAX_PATIENT, verbose=1),
          ModelCheckpoint(filepath=best_model_file, monitor='loss', verbose=1, save_best_only=True)]

#compile the model
model.compile(optimizer="adam",loss="binary_crossentropy",metrics=['accuracy'])

#train
print('training started.... please wait!')

history = model.fit(x=X_train, y=y_train,
                    epochs=MAX_EPOCHS,
                    batch_size=MAX_BATCH_SIZE,
                    verbose=0,
                    validation_data= (X_val, y_val),
                    callbacks=callback,
                    class_weight=Class_Weights )
print('training finished')
#
# quick evaludate model
print('Evaluate model with test data')
score = model.evaluate(x=X_test,y=y_test)

print('test loss:', score[0])
print('test accuracy:', score[1])
################################################################################

%%time
#%%memit

import matplotlib.pyplot as plt
#Plot loss and accuracy for the training and validation set.
def plot_history(history):
    loss_list = [s for s in history.history.keys() if 'loss' in s and 'val' not in s]
    val_loss_list = [s for s in history.history.keys() if 'loss' in s and 'val' in s]
    acc_list = [s for s in history.history.keys() if 'acc' in s and 'val' not in s]
    val_acc_list = [s for s in history.history.keys() if 'acc' in s and 'val' in s]
    if len(loss_list) == 0:
        print('Loss is missing in history')
        return
    plt.figure(figsize=(22,10))
    ## Loss graph
    epochs = range(1,len(history.history[loss_list[0]]) + 1)

    ## Accuracy
    plt.figure(221, figsize=(20,10))
    ## Accuracy
    plt.subplot(221, title='Accuracy')
    for l in acc_list:
        plt.plot(epochs, history.history[l], 'b', label='Training accuracy (' + str(format(history.history[l][-1],'.5f'))+')')
    for l in val_acc_list:
        plt.plot(epochs, history.history[l], 'g', label='Validation accuracy (' + str(format(history.history[l][-1],'.5f'))+')')
    plt.title('Accuracy')
    plt.xlabel('Epochs')
    plt.ylabel('Accuracy')
    plt.legend()
    ## Loss
    plt.subplot(222, title='Loss')
    for l in loss_list:
        plt.plot(epochs, history.history[l], 'b', label='Training loss (' + str(str(format(history.history[l][-1],'.5f'))+')'))
    for l in val_loss_list:
        plt.plot(epochs, history.history[l], 'g', label='Validation loss (' + str(str(format(history.history[l][-1],'.5f'))+')'))
    plt.title('Loss')
    plt.xlabel('Epochs')
    plt.ylabel('Loss')
    plt.legend()
    plt.show()

# plot history
plot_history(history)
################################################################################

#Evaludate model use Keras reported accuracy:
score = model.evaluate(X_train, y_train, verbose=0)
print ("model train data score       : ",round(score[1]*100) , "%")

score = model.evaluate(X_test, y_test, verbose=0)
print ("model test data score        : ",round(score[1]*100) , "%")

score = model.evaluate(X_val, y_val, verbose=0)
print ("model validation data score  : ", round(score[1]*100), "%")

################################################################################

print(list_labels)



from sklearn.metrics import classification_report
import numpy as np


y_true = np.argmax(y_test, axis = 1)
X_valid = np.expand_dims(X_test, axis = -1)
y_pred = model.predict(X_test)
y_pred = np.argmax(y_pred, axis = 1)
labels = [0,1]
target_names = list_labels

print(y_true.shape, y_pred.shape)
print(classification_report(y_true, y_pred, target_names=target_names))


# Prediction test
print ("Prediction with [train] data")
y_pred = model.predict_classes(X_train)
missed=[]
matched=[]
for i in range(len(y_pred)):
    y_val_label_int = argmax(y_train[i])
    if (y_pred[i]!=y_val_label_int):
        missed.append( (y_pred[i], "-", int_to_label[y_pred[i]], " - ", int_to_label[y_val_label_int] ))
    else:
        matched.append((y_pred[i], "-", int_to_label[y_pred[i]], " - ", int_to_label[y_val_label_int]))

print ("  --> match    :", len(matched))
print ("  --> miss     :", len(missed))
print ("  --> accuracy :", round((len(matched)-len(missed))/len(matched)*100,2), "%")
print ("")
#print ("Value missed : \n",missed)

# show sample results
print ("---samples---")
for i in range(220):
    if int_to_label[y_pred[i]] != int_to_label[argmax(y_train[i])]:
        print('predict', int_to_label[y_pred[i]])
        print('original', int_to_label[argmax(y_train[i])])
        print ("")


num_=10
print("Training complete...")
np.set_printoptions(precision=2,suppress=True)
for i in range(len(X_val)):
    print("Prediction: " + str(model.predict([X_val[i].reshape(1,X_val[i].shape[0],126,1)])) + " | True: " + str(y_val[i])) # MÅ FÅ DEN PÅ EN LINJE


# prediction class
print ("Prediction with [test] data")
y_pred = model.predict_classes(X_test)
missed=[]
matched=[]
for i in range(len(y_pred)):
    y_val_label_int = argmax(y_test[i])
    if (y_pred[i]!=y_val_label_int):
        missed.append( (y_pred[i], "-", int_to_label[y_pred[i]], " - ", int_to_label[y_val_label_int] ))
    else:
        matched.append((y_pred[i], "-", int_to_label[y_pred[i]], " - ", int_to_label[y_val_label_int]))

print ("  --> match    :", len(matched))
print ("  --> miss     :", len(missed))
print ("  --> accuracy :", round((len(matched)-len(missed))/len(matched)*100,2), "%")
print ("")

# show sample results
print ("---samples---")
for i in range(62):
    print(i,"predict =", int_to_label[y_pred[i]])
    print (i,"original=", int_to_label[argmax(y_test[i])])
    print ("")

# prediction class
print ("Prediction with [validation] data")
y_pred = model.predict_classes(X_val)
missed=[]
matched=[]
for i in range(len(y_pred)):
    y_val_label_int = argmax(y_val[i])
    if (y_pred[i]!=y_val_label_int):
        missed.append( (y_pred[i], "-", int_to_label[y_pred[i]], " - ", int_to_label[y_val_label_int] ))
    else:
        matched.append((y_pred[i], "-", int_to_label[y_pred[i]], " - ", int_to_label[y_val_label_int]))

print ("  |__match    :", len(matched))
print ("  |__miss     :", len(missed))
print ("  |__accuracy :", round((len(matched)-len(missed))/len(matched)*100,2), "%")
print ("")

# show sample results
print ("---samples---")
for i in range(49):
    print (i,"predict =", int_to_label[y_pred[i]])
    print (i,"original=", int_to_label[argmax(y_val[i])])
    print ("")


predict_array = []
for i in range(len(X_val)):
    predict_array.append(model.predict([X_val[i].reshape(1,X_val[i].shape[0],126,1)]))

print(predict_array[3][0,0])
print(y_val[:,1])


y_val_pos = y_val[:,1]
y_pred_pos = []
for i in range(len(y_val)):
    y_pred_pos.append(predict_array[i][(0,1)])
print(y_pred_pos)


# Get the TPR and FPR for the validation dataset
fpr_keras, tpr_keras, thresholds_keras = roc_curve(y_val_pos, y_pred_pos)

roc_auc = metrics.auc(fpr_keras, tpr_keras)

# method I: plt
import matplotlib.pyplot as plt
plt.title('Receiver Operating Characteristic')
plt.plot(fpr_keras, tpr_keras, 'b', label = 'AUC = %0.2f' % roc_auc)
plt.legend(loc = 'lower right')
plt.plot([0, 1], [0, 1],'r--')
plt.xlim([0, 1])
plt.ylim([0, 1])
plt.ylabel('True Positive Rate')
plt.xlabel('False Positive Rate')
plt.show()

# AUC close to 1 means it has good measure of separability.
prediction = model.predict_classes(X_val)
#metrics.confusion_matrix(y_val_pos, prediction)
tn, fp, fn, tp = confusion_matrix(y_val_pos, prediction).ravel()
print('true negative:',tn,'|','false positive:', fp, '|','false negative:', fn,'|', 'true positive:', tp)


y_train_pos = y_train[:,1]

# AUC close to 1 means it has good measure of separability.
prediction = model.predict_classes(X_train)
#metrics.confusion_matrix(y_val_pos, prediction)
tn, fp, fn, tp = confusion_matrix(y_train_pos, prediction).ravel()
print('true negative:',tn,'|','false positive:', fp, '|','false negative:', fn,'|', 'true positive:', tp)


# prediction class
print ("Prediction with [test] data")
y_pred = model.predict_classes(test_input)
missed=[]
matched=[]

# show sample results
print ("---samples---")
for i in range(30):
    print (i,"predict =", int_to_label[y_pred[i]])
    print ("")
